@@ -1,21 +1,29 @@
 import click
 
-from gretel_client.config import (
-    GretelClientConfigurationError,
-    RunnerMode,
-    configure_session,
-    get_session_config,
-    write_config,
-    ClientConfig,
-)
-from gretel_client.cli.common import SessionContext, pass_session
+from gretel_client.cli.agent import agent
+from gretel_client.cli.artifacts import artifacts
+from gretel_client.cli.common import pass_session, SessionContext
+from gretel_client.cli.connectors import connectors
+from gretel_client.cli.errors import ExceptionHandler
 from gretel_client.cli.models import models
 from gretel_client.cli.projects import projects
 from gretel_client.cli.records import records
-from gretel_client.cli.artifacts import artifacts
+from gretel_client.config import (
+    ClientConfig,
+    configure_session,
+    get_session_config,
+    RunnerMode,
+    write_config,
+)
+from gretel_client.projects.common import f
+from gretel_client.users import users
 
 
-@click.group()
+class GretelCliHandler(ExceptionHandler):
+    ...
+
+
+@click.group(cls=GretelCliHandler)
 @click.option("--debug/--no-debug", default=False, help="Show extra debug messages.")
 @click.option(
     "--output",
@@ -24,11 +32,17 @@ from gretel_client.cli.artifacts import artifacts
 )
 @click.pass_context
 def cli(ctx: click.Context, debug: bool, output: str):
-    """The Gretel CLI.
-    """
+    """The Gretel CLI."""
     ctx.obj = SessionContext(ctx, output_fmt=output, debug=debug)
 
 
+def _check_endpoint(endpoint: str) -> str:
+    endpoint = endpoint.strip("/")
+    if not endpoint.startswith("https://"):
+        endpoint = f"https://{endpoint.split('/')[-1]}"
+    return endpoint
+
+
 @cli.command(help="Configures the Gretel CLI.")
 @click.option(
     "--endpoint",
@@ -41,7 +55,9 @@ def cli(ctx: click.Context, debug: bool, output: str):
     "--default-runner",
     prompt="Default Runner",
     default=lambda: get_session_config().default_runner,
-    type=click.Choice([RunnerMode.CLOUD.value, RunnerMode.LOCAL.value], case_sensitive=False),
+    type=click.Choice(
+        [RunnerMode.CLOUD.value, RunnerMode.LOCAL.value], case_sensitive=False
+    ),
     metavar="RUNNER",
     help="Specify the default runner.",
 )
@@ -65,7 +81,7 @@ def configure(
     sc: SessionContext, endpoint: str, api_key: str, project: str, default_runner: str
 ):
     project_name = None if project == "none" else project
-
+    endpoint = _check_endpoint(endpoint)
     # swap the api key back to the original if one was
     # already read in from the config and not updated
     if api_key.endswith("****"):
@@ -73,28 +89,28 @@ def configure(
     config = ClientConfig(
         endpoint=endpoint, api_key=api_key, default_runner=default_runner
     )
-
-    try:
-        config.update_default_project(project_id=project_name)
-    except GretelClientConfigurationError as ex:
-        sc.log.error(f"The project {project_name} is invalid", ex=ex)
-        sc.exit(1)
-
+    config.update_default_project(project_id=project_name)
     configure_session(config)
 
-    try:
-        config_path = write_config(config)
-        sc.log.info(f"Configuration written to {config_path}. Done.")
-    except Exception as ex:
-        sc.log.error("Could not write configuration to.", ex=ex)
+    config_path = write_config(config)
+    sc.log.info(f"Configuration written to {config_path}. Done.")
 
     sc.print(data=config.masked)
 
 
+@cli.command(help="Check account, user and configuration details.")
+@pass_session
+def whoami(sc: SessionContext):
+    me = users.get_me()
+    sc.print(data={f.EMAIL: me[f.EMAIL], "config": sc.config.masked})
+
+
 cli.add_command(models)
 cli.add_command(records)
 cli.add_command(projects)
 cli.add_command(artifacts)
+cli.add_command(connectors)
+cli.add_command(agent)
 
 
 if __name__ == "__main__":
