@@ -1,27 +1,21 @@
 import json
 import signal
-import sys
-import traceback
+
 from pathlib import Path
-from typing import Any, Callable, Dict, Optional, Union
-from urllib.parse import urlparse
+from typing import Any, Callable, Optional, Tuple, Union
 
 import click
-import requests
 
-from gretel_client.config import (
-    RunnerMode,
-    configure_custom_logger,
-    get_session_config,
-)
-from gretel_client.projects.common import ModelArtifact, ModelType, WAIT_UNTIL_DONE
-from gretel_client.projects.docker import DockerEnvironmentError, check_docker_env
+from gretel_client.cli.utils.parser_utils import RefData
+from gretel_client.config import configure_custom_logger, get_session_config, RunnerMode
+from gretel_client.models.config import get_status_description, StatusDescriptions
+from gretel_client.projects.common import WAIT_UNTIL_DONE
+from gretel_client.projects.docker import check_docker_env
+from gretel_client.projects.exceptions import DockerEnvironmentError
 from gretel_client.projects.jobs import Job, WaitTimeExceeded
 from gretel_client.projects.models import Model
-from gretel_client.projects.projects import Project, get_project
-from gretel_client.projects.records import RecordHandler
-from gretel_client.rest.exceptions import ApiException, NotFoundException
-
+from gretel_client.projects.projects import get_project, Project
+from gretel_client.rest.exceptions import ApiException
 
 ExT = Union[str, Exception]
 
@@ -44,13 +38,13 @@ section G paragraph 2.
 
 
 class Logger:
-    """This classed is used to print CLI progress a debug messages
+    """This class is used to print CLI progress and debug messages
     to the console.
 
-    You will note that all messages are printed out to ``stderr``. This is by
-    design. All progress and debug messages go to ``stderr`` and any ``Response``
-    type classes go to ``stdout``. This keeps ``stdout`` clean so that output
-    can be piped and parsed by downstream commands.
+    All messages are printed out to ``stderr``. All progress and debug messages
+    go to ``stderr`` and any response type data goes to ``stdout``. This keeps
+    ``stdout`` clean so that output may be piped and parsed by downstream
+    commands.
     """
 
     def __init__(self, debug: bool = False):
@@ -60,7 +54,7 @@ class Logger:
         if obj:
             return json.dumps(obj, indent=4)
 
-    def info(self, msg: str = "", data: Optional[Any] = None):
+    def info(self, msg: str = "", data: Optional[Any] = None) -> None:
         """Prints general info statements to the console. Use this log
         level if you want to print messages that indicate progress or
         state change.
@@ -72,8 +66,11 @@ class Logger:
             msg = f"{msg}\n{self._format_object(data)}"
         click.echo(click.style("INFO: ", fg="green") + msg, err=True)
 
-    def warn(self, msg):
-        """Print warn log messages"""
+    def warn(self, msg: str) -> None:
+        self.warning(msg)
+
+    def warning(self, msg: str) -> None:
+        """Prints warn log messages."""
         click.echo(click.style("WARN: ", fg="yellow") + msg, err=True)
 
     def error(self, msg: str = None, ex: ExT = None, include_tb: bool = True):
@@ -88,15 +85,6 @@ class Logger:
         if msg:
             click.echo(click.style("ERROR: ", fg="red") + msg, err=True)
 
-        if include_tb and self.debug_mode:
-            if ex:
-                click.echo(ex, err=True)
-            if include_tb:
-                _, _, tb = sys.exc_info()
-                traceback.print_tb(tb)
-        if ex:
-            self.hint(ex)
-
     def hint(self, ex: ExT):
         hint = None
         try:
@@ -115,9 +103,6 @@ class Logger:
         """
         if self.debug_mode:
             click.echo(click.style("DEBUG: ", fg="blue") + msg, err=True)
-            if ex:
-                _, _, tb = sys.exc_info()
-                traceback.print_tb(tb)
 
 
 def _naming_hint(ex: ExT) -> Optional[str]:
@@ -162,6 +147,9 @@ class SessionContext(object):
     runner: Optional[str] = None
 
     data_source: Optional[str] = None
+    ref_data: Optional[RefData] = None
+
+    model_id: Optional[str] = None
 
     def __init__(self, ctx: click.Context, output_fmt: str, debug: bool = False):
         self.debug = debug
@@ -194,11 +182,16 @@ class SessionContext(object):
             )
             click.echo(_copyright_data + "\n\n", err=True)
 
-    def print(self, *, ok: bool = True, message: str = None, data: Union[list, dict]):
+    def print(
+        self, *, ok: bool = True, message: str = None, data: Union[list, dict, str]
+    ):
         if self.output_fmt == "json":
-            click.echo(json.dumps(data, indent=4))
+            if isinstance(data, str):
+                click.echo(data)
+            else:
+                click.echo(json.dumps(data, indent=4))
         else:
-            click.UsageError("Invalid output format", ctx=self.ctx)
+            click.UsageError("Invalid output format.", ctx=self.ctx)
         if not ok:
             self.exit(1)
 
@@ -208,22 +201,15 @@ class SessionContext(object):
     def set_model(self, model_id: str):
         self.model_id = model_id
 
+    def set_record_handler(self, record_handler_id: str):
+        self.record_handler_id = record_handler_id
+
     @property
     def model(self) -> Model:
         if self._model:
             return self._model
-        else:
-            try:
-                self._model = self.project.get_model(self.model_id)
-                return self._model
-            except NotFoundException as ex:
-                self.log.debug(f"Could not get model {self.model_id}", ex=ex)
-                raise click.BadParameter(
-                    (
-                        f"The model `{self.model_id}` was not found in the "
-                        f"project `{self.project.name}`. "
-                    )
-                )
+        self._model = self.project.get_model(self.model_id)
+        return self._model
 
     @property
     def project(self) -> Project:
@@ -231,15 +217,8 @@ class SessionContext(object):
             return self._project
         if not self._project_id:
             raise click.BadArgumentUsage("A project must be specified.")
-        else:
-            try:
-                self._project = get_project(name=self._project_id)
-                return self._project
-            except Exception as ex:
-                self.log.error(ex=ex, include_tb=True)
-                raise click.BadArgumentUsage(
-                    f"Could not load the specified project `{self._project_id}`"
-                ) from ex
+        self._project = get_project(name=self._project_id)
+        return self._project
 
     def ensure_project(self):
         if not self.project:
@@ -249,20 +228,20 @@ class SessionContext(object):
 
     def _cleanup(self, sig, frame):
         if self._shutting_down:
-            self.log.warn("Got a second interrupt. Shutting down")
+            self.log.warn("Got a second interrupt. Shutting down.")
             self.exit(1)
         else:
             self._shutting_down = True
-        self.log.warn("Got interupt signal.")
+        self.log.warn("Got interrupt signal.")
 
         if self.cleanup_methods:
-            self.log.warn("Attemping graceful shutdown")
+            self.log.warn("Attempting graceful shutdown.")
             for method in self.cleanup_methods:
                 try:
                     method()
                 except Exception as ex:
-                    self.log.debug("Cleanup hook failed to run", ex=ex)
-        self.log.info("Quitting")
+                    self.log.debug("Cleanup hook failed to run.", ex=ex)
+        self.log.info("Quitting.")
         self.exit(0)
 
     def register_cleanup(self, method: Callable):
@@ -283,7 +262,8 @@ def project_option(fn):
     return click.option(  # type:ignore
         "--project",
         allow_from_autoenv=True,
-        help="Gretel project to execute command from",
+        envvar="GRETEl_DEFAULT_PROJECT",
+        help="Gretel project to execute command from.",
         metavar="NAME",
         callback=callback,
     )(fn)
@@ -293,15 +273,13 @@ def runner_option(fn):
     def callback(ctx, param: click.Option, value: str):
         sc: SessionContext = ctx.ensure_object(SessionContext)
         selected_runner = sc.runner or value
-        if selected_runner == "local":
+        if selected_runner == RunnerMode.LOCAL.value:
             try:
                 check_docker_env()
             except DockerEnvironmentError as ex:
-                sc.log.error(
-                    "Runner is local, but docker is not running. Please check that docker is installed and running.",
-                    ex=ex,
-                )
-                sc.exit(1)
+                raise DockerEnvironmentError(
+                    "Runner is local, but docker is not running. Please check that docker is installed and running."
+                ) from ex
         return selected_runner
 
     return click.option(  # type: ignore
@@ -315,22 +293,69 @@ def runner_option(fn):
     )(fn)
 
 
-def model_option(fn):
-    def callback(ctx, param: click.Option, value: str):
-        model_obj = ModelObjectReader(value)
-        sc: SessionContext = ctx.ensure_object(SessionContext)
+def get_model(ctx, param: click.Option, value: str):
+    model_obj = ModelObjectReader(value)
+    sc: SessionContext = ctx.ensure_object(SessionContext)
+    if not sc.model_id:
         model_obj.apply(sc)
-        return sc.model_id or value
+    return sc.model_id or value
 
+
+def model_option(fn):
     return click.option(  # type:ignore
         "--model-id",
         metavar="UID",
         help="Specify the model.",
+        callback=get_model,
+        required=True,
+    )(fn)
+
+
+def record_handler_option(fn):
+    def callback(ctx, param: click.Option, value: str):
+        record_handler_obj = JobObjectReader(value)
+        sc: SessionContext = ctx.ensure_object(SessionContext)
+        record_handler_obj.apply(sc)
+        return sc.record_handler_id or value
+
+    return click.option(  # type:ignore
+        "--record-handler-id",
+        metavar="UID",
+        help="Specify the record handler id.",
         callback=callback,
         required=True,
     )(fn)
 
 
+def ref_data_option(fn):
+    def callback(ctx, param: click.Option, value: Union[RefData, Tuple[str]]):
+        gc: SessionContext = ctx.ensure_object(SessionContext)
+        return value or gc.ref_data
+
+    return click.option(
+        "--ref-data",
+        metavar="PATH",
+        multiple=True,
+        callback=callback,
+        help="Specify additional model or record handler reference data.",
+    )(fn)
+
+
+def parse_file(val: str) -> dict:
+    contents = val
+    obj = {}
+    try:
+        if Path(val).is_file():
+            contents = Path(val).read_text()
+    except OSError:
+        pass
+    try:
+        obj = json.loads(contents)
+    except Exception:
+        pass
+    return obj
+
+
 class ModelObjectReader:
     """Reads a model config and configures the ``SessionContext`` based
     on the contents of the model.
@@ -338,21 +363,7 @@ class ModelObjectReader:
 
     def __init__(self, input: str):
         self.input = input
-        self.model = self._maybe_parse_model(input)
-
-    def _maybe_parse_model(self, val: str) -> dict:
-        contents = val
-        model_obj = {}
-        try:
-            if Path(val).is_file():
-                contents = Path(val).read_text()
-        except OSError:
-            pass
-        try:
-            model_obj = json.loads(contents)
-        except Exception:
-            pass
-        return model_obj
+        self.model = parse_file(input)
 
     def apply(self, sc: SessionContext):
         model_id = self.model.get("uid")
@@ -362,8 +373,7 @@ class ModelObjectReader:
             sc.set_project(project_id)
         if model_id:
             sc.set_model(model_id)
-            sc.data_source = sc.model.data_source
-        if not model_id:
+        else:
             # if there isn't a model id, then we implicitly assume
             # the original input was a model id rather than a model
             # object.
@@ -374,84 +384,31 @@ class ModelObjectReader:
             )
 
 
-StatusDescriptions = Dict[str, Dict[str, str]]
+class JobObjectReader:
+    """Reads a record handler config and configures the ``SessionContext`` based
+    on the contents of the record handler."""
 
-model_create_status_descriptions: StatusDescriptions = {
-    "created": {
-        "default": "Model creation has been queued.",
-    },
-    "pending": {
-        "default": "A worker is being allocated to begin model creation.",
-        "cloud": "A Gretel Cloud worker is being allocated to begin model creation.",
-        "local": "A local container is being started to begin model creation.",
-    },
-    "active": {
-        "default": "A worker has started creating your model!",
-    },
-}
+    def __init__(self, input: str):
+        self.input = input
+        self.record_handler = parse_file(input)
 
-record_generation_status_descriptions: StatusDescriptions = {
-    "created": {
-        "default": "A Record generation job has been queued.",
-    },
-    "pending": {
-        "default": "A worker is being allocated to begin generating synthetic records.",
-        "cloud": "A Gretel Cloud worker is being allocated to begin generating synthetic records.",
-        "local": "A local container is being started to begin record generation.",
-    },
-    "active": {
-        "default": "A worker has started!",
-    },
-}
-
-record_transform_status_descriptions: StatusDescriptions = {
-    "created": {
-        "default": "A Record transform job has been queued.",
-    },
-    "pending": {
-        "default": "A worker is being allocated to begin running a transform pipeline.",
-        "cloud": "A Gretel Cloud worker is being allocated to begin transforming records.",
-        "local": "A local container is being started to and will begin transforming records.",
-    },
-    "active": {
-        "default": "A worker has started!",
-    },
-}
-
-record_classify_status_descriptions: StatusDescriptions = {
-    "created": {
-        "default": "A Record classify job has been queued.",
-    },
-    "pending": {
-        "default": "A worker is being allocated to begin running a classification pipeline.",
-        "cloud": "A Gretel Cloud worker is being allocated to begin classifying records.",
-        "local": "A local container is being started and will begin classifying records.",
-    },
-    "active": {
-        "default": "A worker has started!",
-    },
-}
-
-
-def get_description_set(job: Job) -> Optional[dict]:
-    if isinstance(job, Model):
-        return model_create_status_descriptions
-    if isinstance(job, RecordHandler):
-        if job.model_type == ModelType.SYNTHETICS:
-            return record_generation_status_descriptions
-        if job.model_type == ModelType.TRANSFORMS:
-            return record_transform_status_descriptions
-        if job.model_type == ModelType.CLASSIFY:
-            return record_classify_status_descriptions
-
-
-def get_status_description(
-    descriptions: StatusDescriptions, status: str, runner: str
-) -> str:
-    status_desc = descriptions.get(status)
-    if not status_desc:
-        return ""
-    return status_desc.get(runner, status_desc.get("default", ""))
+    def apply(self, sc: SessionContext):
+        record_handler_id = self.record_handler.get("uid")
+        model_id = self.record_handler.get("model_id")
+        project_id = self.record_handler.get("project_id")
+        runner = self.record_handler.get("runner_mode")
+        if record_handler_id:
+            sc.set_record_handler(record_handler_id)
+        else:
+            sc.set_record_handler(self.input)
+        if project_id:
+            sc.set_project(project_id)
+        if model_id:
+            sc.set_model(model_id)
+        if runner:
+            sc.runner = (
+                RunnerMode.CLOUD.value if runner == "cloud" else RunnerMode.LOCAL.value
+            )
 
 
 def poll_and_print(
@@ -491,29 +448,3 @@ def poll_and_print(
                 f"Exiting the script, but the job will remain running until it reaches the end state."
             )
         sc.exit(0)
-
-
-def download_artifacts(sc: SessionContext, output: str, job: Job):
-    output_path = Path(output)
-    output_path.mkdir(exist_ok=True, parents=True)
-    sc.log.info(f"Downloading model artifacts to {output_path.resolve()}")
-    for artifact_type, download_link in job.get_artifacts():
-        # we don't need to download cloud model artifacts
-        if (
-            isinstance(job, Model)
-            and job.is_cloud_model
-            and artifact_type == ModelArtifact.MODEL.value
-        ):
-            continue
-        try:
-            art = requests.get(download_link)
-            if art.status_code == 200:
-                art_output_path = output_path / Path(urlparse(download_link).path).name
-                with open(art_output_path, "wb+") as out:
-                    sc.log.info(f"\tWriting {artifact_type} to {art_output_path}")
-                    out.write(art.content)
-        except requests.exceptions.HTTPError as ex:
-            sc.log.error(
-                f"\tCould not download {artifact_type}. You might retry this request.",
-                ex=ex,
-            )
