@@ -1,8 +1,12 @@
 """
 High level API for interacting with a Gretel Project
 """
+import json
+import uuid
+
 from contextlib import contextmanager
 from functools import wraps
+from io import BytesIO
 from pathlib import Path
 from typing import Any, Dict, Iterator, List, Optional, Type, TypeVar, Union
 from urllib.parse import urlparse
@@ -10,13 +14,22 @@ from urllib.parse import urlparse
 import requests
 import smart_open
 
-from gretel_client.config import get_session_config
-from gretel_client.projects.common import f, validate_data_source
+from tenacity import retry, retry_if_exception_type, stop_after_attempt, wait_fixed
+
+from gretel_client.cli.utils.parser_utils import (
+    DataSourceTypes,
+    ref_data_factory,
+    RefDataTypes,
+)
+from gretel_client.config import get_logger, get_session_config
+from gretel_client.projects.common import _DataFrameT, f, validate_data_source
+from gretel_client.projects.exceptions import GretelProjectError
 from gretel_client.projects.models import Model
 from gretel_client.rest import models
 from gretel_client.rest.api.projects_api import ProjectsApi
-from gretel_client.rest.exceptions import UnauthorizedException
+from gretel_client.rest.exceptions import NotFoundException, UnauthorizedException
 from gretel_client.rest.model.artifact import Artifact
+from gretel_client.users.users import get_me
 
 DATA = "data"
 PROJECTS = "projects"
@@ -24,8 +37,7 @@ PROJECT = "project"
 MODELS = "models"
 
 
-class GretelProjectError(Exception):
-    ...
+log = get_logger(__name__)
 
 
 def check_not_deleted(func):
@@ -43,6 +55,19 @@ def check_not_deleted(func):
 MT = TypeVar("MT", dict, Model)
 
 
+# These exceptions are used for control flow with retries in get_artifact_manifest.
+# They are NOT intended to bubble up out of this module.
+class ManifestNotFoundException(Exception):
+    pass
+
+
+class ManifestPendingException(Exception):
+    def __init__(self, msg: str = None, manifest: dict = {}):
+        # Piggyback the pending manifest onto the exception.  If we give up, we still want to pass it back as a normal return value.
+        self.manifest = manifest
+        super().__init__(msg)
+
+
 class Project:
     """Represents Gretel project. In general you should not have
     to init this class directly, but can make use of the factory
@@ -69,17 +94,18 @@ class Project:
         self._deleted = False
 
     @check_not_deleted
-    def delete(self, include_models: bool = True):
+    def delete(self, *args, **kwargs):
         """Deletes a project. After the project has been deleted, functions
         relying on a project will fail with a ``GretelProjectError`` exception.
 
         Note: Deleting projects is asynchronous. It may take a few seconds
         for the project to be deleted by Gretel services.
         """
-        if include_models:
-            for model in self.search_models():
-                model.delete()
-
+        # todo(dn): remove in 0.9 release
+        if kwargs.get("include_models") or len(args) > 0:
+            log.warning(
+                "``include_models`` is deprecated and will be removed in the 0.9 release."
+            )
         self.projects_api.delete_project(project_id=self.project_id)
         self._deleted = True
 
@@ -106,7 +132,9 @@ class Project:
         return self.projects_api.get_project(project_id=self.name).get(DATA)
 
     @check_not_deleted
-    def search_models(self, factory: Type[MT] = Model, limit: int = 100) -> Iterator[MT]:
+    def search_models(
+        self, factory: Type[MT] = Model, limit: int = 100
+    ) -> Iterator[MT]:
         """Search for project models.
 
         Args:
@@ -133,13 +161,16 @@ class Project:
         """Lookup and return a Project ``Model`` by it's ``model_id``.
 
         Args:
-            model_id: The ``model_id` to lookup
+            model_id: The ``model_id`` to lookup
         """
         return Model(project=self, model_id=model_id)
 
     @check_not_deleted
     def create_model_obj(
-        self, model_config: Union[str, Path, dict], data_source: Optional[str] = None
+        self,
+        model_config: Union[str, Path, dict],
+        data_source: Optional[DataSourceTypes] = None,
+        ref_data: Optional[RefDataTypes] = None,
     ) -> Model:
         """Creates a new model object. This will not submit the model
         to Gretel's cloud API. Please refer to the ``Model`` docs for
@@ -148,22 +179,30 @@ class Project:
         Args:
             model_config: Specifies a model config. For more information
                 about model configs, please refer to our doc site,
-                https://docs.gretel.ai/model-configurations.
+                https://docs.gretel.ai/reference/model-configurations.
             data_source: Defines the model data_source. If the model_config
                 already has a data_source defined, this property will
                 override the existing one.
+            ref_data: An Optional str, dict, dataframe or list of reference data sources
+                to upload for the job.
         """
         _model = Model(model_config=model_config, project=self)
-        if data_source and not isinstance(data_source, str):
-            raise ValueError("data_source must be a str")
+        if (
+            not isinstance(data_source, _DataFrameT)
+            and data_source
+            and not isinstance(data_source, str)
+        ):
+            raise ValueError("data_source must be a str or dataframe")
         if data_source is not None:
             _model.data_source = data_source
+        ref_data_obj = ref_data_factory(ref_data)
+        _model.ref_data = ref_data_obj
         return _model
 
     @property
     @check_not_deleted
     def artifacts(self) -> List[dict]:
-        """Returns a list of project artifacts"""
+        """Returns a list of project artifacts."""
         return (
             self.projects_api.get_artifacts(project_id=self.name)
             .get(DATA)
@@ -171,23 +210,29 @@ class Project:
         )
 
     def upload_artifact(
-        self, artifact_path: Union[Path, str], _validate: bool = True
+        self, artifact_path: Union[Path, str, _DataFrameT], _validate: bool = True
     ) -> str:
         """Resolves and uploads the data source specified in the
         model config.
 
         Returns:
-            A Gretel artifact key
+            A Gretel artifact key.
         """
         if isinstance(artifact_path, str) and artifact_path.startswith("gretel_"):
             return artifact_path
-        if _validate:
-            validate_data_source(artifact_path)
-        if isinstance(artifact_path, Path):
-            artifact_path = str(artifact_path)
-        with smart_open.open(artifact_path, "rb", ignore_ext=True) as src:  # type:ignore
-            src_data = src.read()
+        if isinstance(artifact_path, _DataFrameT):
+            artifact_path = BytesIO(artifact_path.to_csv(index=False).encode("utf-8"))
+            file_name = f"dataframe-{uuid.uuid4()}.csv"
+        else:
+            if _validate:
+                validate_data_source(artifact_path)
+            if isinstance(artifact_path, Path):
+                artifact_path = str(artifact_path)
             file_name = Path(urlparse(artifact_path).path).name
+        with smart_open.open(
+            artifact_path, "rb", ignore_ext=True
+        ) as src:  # type:ignore
+            src_data = src.read()
             art_resp = self.projects_api.create_artifact(
                 project_id=self.name, artifact=Artifact(filename=file_name)
             )
@@ -197,20 +242,69 @@ class Project:
                 data=src_data,
             )
             if upload_resp.status_code != 200:
-                raise GretelProjectError(f"Could not upload artifact {artifact_path}")
+                upload_resp.raise_for_status()
             return artifact_key
 
     def delete_artifact(self, key: str):
-        """Deletes a project artifact
+        """Deletes a project artifact.
 
         Args:
-            key: Artifact key to delete
+            key: Artifact key to delete.
         """
         return self.projects_api.delete_artifact(project_id=self.name, key=key)
 
+    def get_artifact_link(self, key: str) -> str:
+        """Returns a link to download a project artifact.
+
+        Args:
+            key: Project artifact key to generate download url for.
+
+        Returns:
+            A signed URL that may be used to download the given
+            project artifact.
+        """
+        resp = self.projects_api.download_artifact(project_id=self.name, key=key)
+        return resp[f.DATA][f.DATA][f.URL]
+
+    # The server side API will return manifests with PENDING status if artifact processing has not completed
+    # or it will return a 404 (not found) if you immediately request the artifact before processing has even started.
+    # This is correct but not convenient.  To keep every end user from writing their own retry logic, we add some here.
+    @retry(
+        wait=wait_fixed(2),
+        stop=stop_after_attempt(15),
+        retry=retry_if_exception_type(ManifestPendingException),
+        # Instead of throwing an exception, return the pending manifest.
+        retry_error_callback=lambda retry_state: retry_state.outcome.exception().manifest,
+    )
+    @retry(
+        wait=wait_fixed(3),
+        stop=stop_after_attempt(5),
+        retry=retry_if_exception_type(ManifestNotFoundException),
+        # Instead of throwing an exception, return None.  Given that we waited for a short grace period to let the artifact become PENDING,
+        # if we are still getting 404's the key probably does not actually exist.
+        retry_error_callback=lambda retry_state: None,
+    )
+    def get_artifact_manifest(
+        self, key: str, retry_on_not_found: bool = True, retry_on_pending: bool = True
+    ) -> dict:
+        _path = f"/projects/{self.name}/artifacts/manifest"
+        _query_params = {"key": key}
+        resp = None
+        try:
+            resp = self.projects_api.get_artifact_manifest(
+                project_id=self.name, key=key
+            )
+        except NotFoundException:
+            raise ManifestNotFoundException()
+        if retry_on_not_found and resp is None:
+            raise ManifestNotFoundException()
+        if retry_on_pending and resp is not None and resp.get("status") == "pending":
+            raise ManifestPendingException(manifest=resp)
+        return resp
+
 
 def search_projects(limit: int = 200, query: str = None) -> List[Project]:
-    """Searches for project
+    """Searches for project.
 
     Args:
         limit: The max number of projects to return.
@@ -318,10 +412,10 @@ def get_project(
                 resp = api.create_project(project=models.Project(**project_args))
                 project = api.get_project(project_id=resp.get(DATA).get("id"))
             else:
-                raise
+                raise GretelProjectError(f"Could not get project using '{name}'.")
 
     if not project:
-        raise GretelProjectError("Could not get or create project.")
+        raise GretelProjectError(f"Could not get or create project using '{name}'.")
 
     p = project.get(DATA).get(PROJECT)
 
@@ -340,6 +434,7 @@ def tmp_project():
     The project will be deleted from Gretel Cloud when the scope is exited.
 
     Example::
+
         with tmp_project() as proj:
             model = proj.create_model_obj()
     """
@@ -348,3 +443,47 @@ def tmp_project():
         yield project
     finally:
         project.delete()
+
+
+def create_or_get_unique_project(
+    *, name: str, desc: str = None, display_name: str = None
+) -> Project:
+    """
+    Helper function that provides a consistent experience for creating
+    and fetching a project with the same name. Given a name of a project,
+    this helper will fetch the current user's ID and use that as a suffix
+    in order to create a project name unique to that user. Once the project
+    is created, if it can be fetched, it will not be re-created over and
+    over again.
+
+    Args:
+        name: The name of the project, which will have the User's ID appended
+            to it automatically.
+        desc: Description of the project.
+        display_name: If None, the display name will be set equal to the value
+            of ``name`` _before_ the user ID is appended.
+
+    NOTE:
+        The ``desc`` and ``display_name`` parameters will only be used when
+        the project is first created. If the project already exists, these
+        params will have no affect.
+    """
+    current_user_dict = get_me()
+    unique_suffix = current_user_dict["_id"][9:]
+    target_name = f"{name}-{unique_suffix}"
+
+    try:
+        project = get_project(name=target_name)
+        return project
+    except GretelProjectError:
+        # Project name not found
+        pass
+    except Exception:
+        raise
+
+    # Try and create the project if we coud not find it
+    # originally
+    project = create_project(
+        name=target_name, display_name=display_name or name, desc=desc
+    )
+    return project
