@@ -1,13 +1,16 @@
 import json
+import tempfile
+
 from pathlib import Path
 from typing import Callable, List
-from unittest.mock import MagicMock
-import tempfile
+from unittest.mock import MagicMock, patch
 
 import pytest
 import yaml
 
+from gretel_client.cli.utils.parser_utils import RefData
 from gretel_client.config import RunnerMode
+from gretel_client.projects.exceptions import DataSourceError
 from gretel_client.projects.models import Model, ModelConfigError, read_model_config
 
 
@@ -154,7 +157,10 @@ def test_model_submit_no_local_mode(m: Model):
     assert "local" in str(err)
 
 
-def test_does_poll_status_and_logs(m: Model, model_logs: List[dict]):
+@patch("time.sleep")
+def test_does_poll_status_and_logs(
+    sleep_patch: MagicMock, m: Model, model_logs: List[dict]
+):
     m._submit(runner_mode=RunnerMode.LOCAL, _default_manual=True)
     m._projects_api.get_model.side_effect = [  # type:ignore
         {"data": {"model": {"status": "created"}}},
@@ -185,12 +191,13 @@ def test_does_read_model_short_path():
 
 
 def test_does_not_read_bad_local_data():
-    with tempfile.NamedTemporaryFile() as tmp_config:
+    with tempfile.NamedTemporaryFile(delete=False) as tmp_config:
         tmp_config.write(b"\tfoo")  # a regular string loads as YAML
         tmp_config.seek(0)
         with pytest.raises(ModelConfigError) as err:
             read_model_config(tmp_config.name)
         assert "YAML or JSON" in str(err)
+    tmp_config.close()
 
 
 def test_does_read_in_memory_model(transform_model_path: Path):
@@ -252,3 +259,94 @@ def test_synth_report_output(m: Model, get_fixture: Callable):
     ]
     for field in expected_fields:
         assert field in peek.keys()
+
+
+def test_xf_report_summary(m: Model, get_fixture: Callable):
+    report_json = get_fixture("xf_report_json.json.gz")
+    m.model_config["models"][0] = {
+        "transforms": {}
+    }  # pretend the model stub is a transforms model
+    summary = m.get_report_summary(str(report_json))
+    # PROD-76, new summary format
+    assert len(summary) == 1 and "summary" in summary
+    expected_fields = {
+        "training_time_seconds",
+        "record_count",
+        "field_count",
+        "field_transforms",
+        "value_transforms",
+        "warnings",
+    }
+    found_fields = {d["field"] for d in summary["summary"]}
+    assert expected_fields == found_fields
+
+
+def test_synth_report_summary(m: Model, get_fixture: Callable):
+    report_json = get_fixture("synth_report_json.json.gz")
+    m.model_config["models"][0] = {
+        "synthetics": []
+    }  # pretend the model stub is a synthetics model
+    summary = m.get_report_summary(str(report_json))
+    # PROD-76, new summary format
+    assert len(summary) == 1 and "summary" in summary
+    expected_fields = {
+        "synthetic_data_quality_score",
+        "field_correlation_stability",
+        "principal_component_stability",
+        "field_distribution_stability",
+    }
+    found_fields = {d["field"] for d in summary["summary"]}
+    assert expected_fields == found_fields
+
+
+def test_can_name_model(m: Model):
+    assert m.name == "my-awesome-model"
+    new_name = "my-model"
+    m.name = new_name
+    assert m.name == new_name
+    assert m.model_config["name"] == new_name
+    assert m._local_model_config["name"] == new_name
+
+
+def test_ref_data(m: Model, transform_local_data_source: Path):
+    # Check non-existent ref data from config
+    assert m.ref_data.is_empty
+
+    # Cloud artifacts
+    ref_data = RefData({"foo": "gretel_abc"})
+    m.ref_data = ref_data
+    assert m.ref_data == ref_data
+    assert not m.external_ref_data
+
+    # "Local file" but cannot verify its location on disk
+    ref_data = RefData({"foo": "bar.csv"})
+    m.ref_data = ref_data
+    assert m.ref_data == ref_data
+    assert m.external_ref_data
+    with pytest.raises(DataSourceError):
+        m.validate_ref_data()
+
+    # Local file seated along with the config
+    ref_data = RefData({"foo": transform_local_data_source.name})
+    m.ref_data = ref_data
+    check = m.ref_data.ref_dict["foo"]
+    assert check.startswith(str(m._local_model_config_path.parent))
+    assert m.external_ref_data
+    m.validate_ref_data()
+
+
+def test_does_write_artifacts_to_disk(tmpdir: Path, m: Model):
+    base_endpoint = (
+        "https://gretel-public-website.s3.us-west-2.amazonaws.com/tests/client/"
+    )
+    files = ["account-balances.csv", "report_json.json.gz", "model.tar.gz"]
+    keys = ["data_preview", "report_json", "model"]
+    m.get_artifacts = MagicMock(
+        return_value=iter(zip(keys, [base_endpoint + f for f in files]))
+    )
+    m.download_artifacts(str(tmpdir))
+    for file in files:
+        if file == "model.tar.gz":
+            assert not (tmpdir / file).exists()
+        else:
+            assert (tmpdir / file).exists()
